#! /usr/bin/env python
"""Command-Line Interface for read_acq."""

from __future__ import annotations

from pathlib import Path

import click

from .gsdata import read_acq_to_gsdata

main = click.Group()


@main.command()
@click.argument(
    "infile",
    nargs=-1,
)
@click.option(
    "-o",
    "--outfile",
    default=None,
    type=click.Path(exists=False, dir_okay=False),
)
@click.option(
    "-d",
    "--direc",
    default=".",
    type=click.Path(exists=True, file_okay=False),
)
@click.option("--telescope", "--loc", default="edges-low", type=str)
@click.option("-n", "--name", default="{year}_{day}", type=str)
def convert(infile, outfile: str | None, direc: str, telescope: str, name: str):
    """Convert an ACQ file to a .gsh5 format file.

    Multiple input files can be provided, and they will be joined together in a single
    output file. Furthermore, the input files can have glob-style patterns in them so
    that they expand to a list of files, e.g.:

        $ acq convert 2023_070*.acq

    If there is more than one file matching the pattern, the files will be sorted, then
    all read and output as a single GSH5 file (by default, the file "2023_070.gsh5").
    """
    infiles = [Path(fl) for fl in infile]

    fls = []
    for fl in infiles:
        fls += sorted(fl.parent.glob(fl.name))

    gsd = read_acq_to_gsdata(fls, telescope=telescope, name=name)
    if outfile is None:
        outfile = f"{gsd.name}.gsh5"

    gsd.write_gsh5(Path(direc) / outfile)



"""Functions that compute coordinates in a fast way.

These functions are equivalent to those used in Alan's C code. They have essentially
just been transcripted from the C-code, and keep the same names as functions there.
"""

from __future__ import annotations

from datetime import datetime
from typing import Union

import numpy as np

Number = Union[float, np.ndarray]


def isleapyear(year: int) -> bool:
    """Check if a year is a leap year."""
    return (year % 4 == 0 and year % 100 != 0) or year % 400 == 0


def gst(time: float) -> float:
    """Convert time to GST."""
    secs = tosecs(2011, 1, 17, 15, 58.0778)
    return (((time - secs) / 86164.09053) % 1) * 2 * np.pi


def datetime_tosecs(dt: datetime):
    """Convert a datetime object to seconds since New Year 1970."""
    return tosecs(
        dt.year,
        dt.timetuple().tm_yday,
        dt.hour,
        dt.minute,
        round(dt.second + dt.microsecond / 1e6),
    )


def tosecs(yr: int, day: int, hour: int, minutes: int, sec: int | float) -> int | float:
    """Convert a date-time to seconds since New Year 1970."""
    secs = (yr - 1970) * 31536000 + (day - 1) * 86400 + hour * 3600 + minutes * 60 + sec

    yr = np.asarray(yr)
    secs = np.asarray(secs)

    for i in range(1970, yr.max()):
        if isleapyear(i):
            secs[yr > i] += 86400

    if secs.size == 1:
        return float(secs)
    else:
        return secs


def galactic_to_radec(glat: Number, glon: Number) -> tuple[Number, Number]:
    """Galactic to radec  2000 epoch pole at 12h51.4 27.1.

    Parameters
    ----------
    glat
        Galactic latitude in degrees.
    glon
        Galactic longitude in degrees.
    """
    deg2rad = np.pi / 180
    glat = glat * deg2rad
    glon = glon * deg2rad  # no in-place so that input is not modified

    d0 = -(28.0 + 56.0 / 60.0) * deg2rad
    r0 = (17.0 + 45.5 / 60.0) * np.pi / 12.0
    dp = 27.1 * np.pi / 180.0
    rp = (12.0 + 51.4 / 60.0) * np.pi / 12.0
    zr = np.sin(d0)
    xr = np.cos(r0 - rp) * np.cos(d0)
    yr = np.sin(r0 - rp) * np.cos(d0)
    xg = xr * np.sin(dp) - zr * np.cos(dp)
    yg = yr
    a = np.arctan2(yg, xg)
    xg = np.cos((glon) + a) * np.cos(glat)
    yg = np.sin((glon) + a) * np.cos(glat)
    zg = np.sin(glat)
    xr = xg * np.sin(dp) + zg * np.cos(dp)
    yr = yg
    zr = zg * np.sin(dp) - xg * np.cos(dp)

    dec = np.arctan2(zr, np.sqrt(xr * xr + yr * yr))
    ra = np.arctan2(yr, xr) + rp
    return ra, dec


def toyrday(secs: float) -> tuple[int, int, int, int, int]:
    """Convert Seconds to Yr/Day/Hr/Min/Sec."""
    day = int(np.floor(secs / 86400.0))
    sec = secs - day * 86400.0
    year = 1970
    while day > 365:
        days = 366 if isleapyear(year) else 365
        day -= days
        year += 1

    hour = int(sec / 3600.0)
    sec -= hour * 3600.0
    minute = int(sec / 60.0)
    sec -= minute * 60
    day += 1

    if day == 366 and not isleapyear(year):
        # fix for alan
        day -= 365
        year += 1

    return year, int(day), int(hour), int(minute), int(sec)


def radec_azel2(
    ha: Number, sindec: Number, cosdec: Number, sinlat: float, coslat: float
) -> tuple[Number, Number]:
    """Convert from RA/DEC to Az/El.

    Parameters
    ----------
    ha
        Hour angle of the observed sources (including time and location information of
        the antenna, as well) in radians.
    sindec
        Sine of the declination of the observation in radians.
    cosdec
        Cosine of the declination of the observation in radians.
    sinlat
        Sine of the latitude of the antenna in radians.
    coslat
        Cosine of the latitude of the antenna in radians.
    """
    w = np.sin(ha) * cosdec
    r = np.cos(ha) * cosdec
    zen = r * coslat + sindec * sinlat
    north = -r * sinlat + sindec * coslat
    el = np.arctan2(zen, np.sqrt(north * north + w * w))
    az = np.arctan2(-w, north) % (2 * np.pi)
    return az, el


def radec_azel(
    ha: Number,
    dec: Number,
    lat: float,
) -> tuple[Number, Number]:
    """Convert from sky to antenna coords az/el.

    Parameters
    ----------
    ha
        Hour angle of the observed sources (including time and location information of
        the antenna, as well) in radians.
    dec
        Declination of the observation in radians.
    lat
        Latitude of the antenna in radians.
    """
    return radec_azel2(ha, np.sin(dec), np.cos(dec), np.sin(lat), np.cos(lat))


def radec_azel_from_lst(
    lst: float, ra: Number, dec: Number, lat: float
) -> tuple[Number, Number]:
    """Convert from sky to antenna coords az/el.

    Parameters
    ----------
    lst
        Local sidereal time in radians.
    ra
        Right ascension of the sources in radians.
    dec
        Declination of the sources in radians.
    lat
        Latitude of the antenna in radians.
    """
    ha = lst - ra
    return radec_azel(ha, dec, lat)



"""Tests of writing the data."""

import os
from pathlib import Path

import numpy as np
import pytest
from pygsdata.select import select_loads

from read_acq import decode_file
from read_acq.gsdata import fast_lst_setter, read_acq_to_gsdata, write_gsdata_to_acq
from read_acq.read_acq import ACQError


@pytest.mark.parametrize("with_fast_lst_setter", [False, True])
def test_roundtrip_gsh5(sample_acq: Path, tmp_path, with_fast_lst_setter):
    _q, _p, _meta = decode_file(sample_acq)
    gsd = read_acq_to_gsdata(
        sample_acq, lst_setter=fast_lst_setter if with_fast_lst_setter else None
    )
    np.testing.assert_allclose(
        _q.T, (gsd.data[0, 0] - gsd.data[1, 0]) / (gsd.data[2, 0] - gsd.data[1, 0])
    )

    write_gsdata_to_acq(gsd, tmp_path / "new.acq")
    q, p, meta = decode_file(tmp_path / "new.acq", progress=False)

    np.testing.assert_allclose(
        q.T,
        (gsd.data[0, 0] - gsd.data[1, 0]) / (gsd.data[2, 0] - gsd.data[1, 0]),
        rtol=3e-2,  # TODO: this is only this high because of MacOS
    )


def test_multifile_read(sample_acq: Path):
    gsd0 = read_acq_to_gsdata(sample_acq)
    gsd = read_acq_to_gsdata([sample_acq, sample_acq])
    assert gsd.ntimes == gsd0.ntimes * 2


def test_non_existent_file():
    with pytest.raises(FileNotFoundError, match="does not exist"):
        read_acq_to_gsdata("non_existent_file.acq")


def test_non_existent_telescope(sample_acq):
    with pytest.raises(
        ValueError, match="telescope must be a Telescope or name of a KNOWN_TELESCOPE"
    ):
        read_acq_to_gsdata(sample_acq, telescope="non_existent_telescope")


def test_write_single_load(sample_acq: Path, tmp_path):
    gsd = select_loads(read_acq_to_gsdata(sample_acq), loads=("ant",))

    with pytest.raises(ValueError, match="Can only encode 3-load data to ACQ file"):
        write_gsdata_to_acq(gsd, tmp_path / "new.acq")


def test_no_files_given():
    with pytest.raises(ValueError, match="No files given to read"):
        read_acq_to_gsdata(path=[])


def test_only_empty_files_given(sample_acq, tmp_path):
    outpath = tmp_path / "empty.acq"

    os.system(f"head -n 31 {sample_acq} > {outpath}")

    gsd = read_acq_to_gsdata([sample_acq])

    with pytest.raises(ACQError, match="No data in any files"):
        read_acq_to_gsdata([outpath])

    with pytest.raises(ACQError, match="No data in any files"):
        read_acq_to_gsdata([outpath, outpath])

    gsd1 = read_acq_to_gsdata([sample_acq, outpath])
    assert gsd1.ntimes == gsd.ntimes

    gsd1 = read_acq_to_gsdata([outpath, sample_acq])
    assert gsd1.ntimes == gsd.ntimes

    gsd1 = read_acq_to_gsdata([outpath, outpath, sample_acq])
    assert gsd1.ntimes == gsd.ntimes

    gsd1 = read_acq_to_gsdata([outpath, sample_acq, outpath])
    assert gsd1.ntimes == gsd.ntimes

    gsd1 = read_acq_to_gsdata([sample_acq, outpath, sample_acq])
    assert gsd1.ntimes == gsd.ntimes * 2



"""Package providing functionality for reading .ACQ format files."""

__all__ = ["decode_file", "encode", "read_acq_to_gsdata", "write_gsdata_to_acq"]

from importlib.metadata import PackageNotFoundError, version

from .gsdata import read_acq_to_gsdata, write_gsdata_to_acq
from .read_acq import decode_file, encode

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError



"""Perform tests of the _coordinates_alan module."""

from datetime import datetime

import numpy as np
from astropy import units as un
from astropy.coordinates import Galactic, SkyCoord

from read_acq import _coordinates as crda


def test_alan_coordinates_toyrday():
    seconds_since_ny1970 = 1474043217.33333333
    year, day, hour, minute, sec = crda.toyrday(seconds_since_ny1970)
    assert year == 2016
    assert day == 260
    assert hour == 16


def test_ra_dec_at_lst():
    az, el = crda.radec_azel_from_lst(lst=0, ra=0, dec=0, lat=0)
    assert az == 0.0
    assert el == np.pi / 2


def test_galactic_to_radec():
    ra = 0.7
    dec = 0.0

    coord = SkyCoord(ra=ra * un.rad, dec=dec * un.rad)
    galframe = Galactic()
    gal = coord.transform_to(galframe)

    _ra, _dec = crda.galactic_to_radec(gal.b.deg, gal.l.deg)
    assert np.isclose(_dec, dec, atol=1e-2)
    assert np.isclose(_ra, ra, atol=1e-2)


def test_non_leap_year_first_day():
    secs = (365 + 365) * 86400 + 1
    year, day, hour, minute, sec = crda.toyrday(secs)
    assert year == 1972
    assert day == 1
    assert hour == 0
    assert minute == 0
    assert sec == 1


def test_datetime_to_secs():
    dt = datetime(year=1970, month=1, day=1, hour=0, minute=0, second=0)
    assert crda.datetime_tosecs(dt) == 0

    dt = datetime(year=1970, month=1, day=1, hour=1, minute=0, second=0)
    assert crda.datetime_tosecs(dt) == 3600

    dt = datetime(year=1970, month=1, day=2, hour=0, minute=0, second=0)
    assert crda.datetime_tosecs(dt) == 86400



"""Tests of the CLI."""

from pathlib import Path

import numpy as np
from click.testing import CliRunner
from pygsdata import GSData

from read_acq import decode_file
from read_acq.cli import convert


def test_convert(tmp_path_factory):
    cli = CliRunner()

    data = Path(__file__).parent / "data/sample.acq"
    outfile = tmp_path_factory.mktemp("direc") / "tempfile.gsh5"
    result = cli.invoke(convert, [str(data), "--outfile", str(outfile)])
    assert result.exit_code == 0

    q, p, meta = decode_file(data)
    gsd = GSData.from_file(outfile)
    assert np.allclose(gsd.data[0, 0].T, p[0])



"""Read an ACQ file into a GSData object."""

from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path

import numpy as np
from astropy import units as un
from astropy.coordinates import EarthLocation, Longitude
from astropy.time import Time
from pygsdata import KNOWN_TELESCOPES, GSData, Telescope
from pygsdata.history import History, Stamp
from pygsdata.readers import gsdata_reader
from pygsdata.utils import time_concat

from . import _coordinates as _crd
from .read_acq import ACQError, decode_file, encode


def fast_lst_setter(times: Time, loc: EarthLocation):
    """Set the LSTs for the GSData object."""
    years, days, hours, minutes, seconds = np.array(
        [yd.split(":") for yd in times.yday.flatten()]
    ).T

    secs = _crd.tosecs(
        years.astype(int),
        days.astype(int),
        hours.astype(int),
        minutes.astype(int),
        seconds.astype(float),
    )
    gst = _crd.gst(secs) * 12 / np.pi + loc.lon.hour

    return Longitude(gst * un.hour).reshape(times.shape)


@gsdata_reader(select_on_read=False, formats=["acq"])
def read_acq_to_gsdata(
    path: str | Path | Sequence[str | Path],
    telescope: Telescope = KNOWN_TELESCOPES["edges-low"],
    name: str = "{year}-{day}:{hour}:{minute}",
    lst_setter: callable | None = None,
    **kwargs,
) -> GSData:
    """Read an ACQ file into a GSData object.

    Parameters
    ----------
    path : str or Path
        The path to the file to read. If a sequence of paths is given, the files will be
        concatenated along the time axis.
    telescope_location : str or EarthLocation
        The location of the telescope.
    name : str
        The name of the GSData object. Can include formatting fields for year, day,
        hour, minute and stem (the stem of the input filename).
    lst_setter : callable, optional
        A function that takes the astropy Time object at which the data is defined and
        returns a set of LSTs that will be stored in the GSData object. By default, use
        the astropy function used by pygsdata itself. Set
    **kwargs
        Additional keyword arguments to pass to the GSData constructor.
    """
    path = [Path(path)] if isinstance(path, (str, Path)) else [Path(p) for p in path]

    if len(path) == 0:
        raise ValueError("No files given to read")

    if not all(p.exists() for p in path):
        raise FileNotFoundError(f"File {path} does not exist")

    if isinstance(telescope, str):
        try:
            telescope = KNOWN_TELESCOPES[telescope]
        except KeyError as e:
            raise ValueError(
                "telescope must be a Telescope or name of a KNOWN_TELESCOPE, "
                f"got {telescope}. Known 'scopes are {KNOWN_TELESCOPES.keys()}."
            ) from e

    # Read the _first_ file to get the metadata
    pants = []
    ploads = []
    plnss = []
    times = []
    ancs = []
    for pth in path:
        _, (pant, pload, plns), anc = decode_file(pth)
        _times = Time(anc.data.pop("times"), format="yday", scale="utc")

        if pant.size == 0:
            continue  # ignore this file, it's empty.

        pants.append(pant)
        ploads.append(pload)
        plnss.append(plns)
        times.append(_times)
        ancs.append(anc)

    if len(pants) == 0:
        raise ACQError(f"No data in any files: {path}")

    pant = np.concatenate(pants, axis=1)
    pload = np.concatenate(ploads, axis=1)
    plns = np.concatenate(plnss, axis=1)
    times = time_concat(times)
    anc = ancs[0]
    anc.data = {k: np.concatenate([a.data[k] for a in ancs]) for k in anc.data}

    year, day, hour, minute = times[0, 0].to_value("yday", "date_hm").split(":")
    name = name.format(year=year, day=day, hour=hour, minute=minute, stem=path[0].stem)

    # TODO: use proper integration time...
    time_ranges = Time(
        np.hstack((times.jd, (times + 13 * un.s).jd)), format="jd"
    ).reshape((*times.shape, 2))

    if lst_setter is not None:
        kwargs["lsts"] = lst_setter(times, telescope.location)
        kwargs["lst_ranges"] = lst_setter(time_ranges, telescope.location)

    return GSData(
        data=np.array([pant.T, pload.T, plns.T])[:, np.newaxis],
        times=times,
        freqs=anc.frequencies * un.MHz,
        data_unit="power",
        loads=("ant", "internal_load", "internal_load_plus_noise_source"),
        auxiliary_measurements={name: anc.data[name] for name in anc.data},
        filename=path[0] if len(path) == 1 else None,
        telescope=telescope,
        name=name,
        history=History(
            (
                Stamp(
                    "Read from ACQ file",
                    function="read_acq_to_gsdata",
                    parameters={"path": path},
                ),
            ),
        ),
        **kwargs,
    )


def write_gsdata_to_acq(
    gsdata: GSData,
    outfile: str | Path,
    temperature: float = 25.0,
    nblk: int = 2974,
):
    """Write a GSData object to an ACQ file.

    Parameters
    ----------
    gsdata : GSData
        The data to write.
    outfile : str or Path
        The file to write to.
    temperature : float
        The temperature of the system, in Celsius.
    nblk : int
        The number of blocks in the data.
    """
    if gsdata.nloads != 3:
        raise ValueError("Can only encode 3-load data to ACQ file.")

    ancillary = {
        "times": gsdata.times.strftime("%Y:%j:%H:%M:%S"),
        "adcmax": gsdata.auxiliary_measurements["adcmax"],
        "adcmin": gsdata.auxiliary_measurements["adcmin"],
    }

    meta = {
        "temperature": temperature,
        "nblk": nblk,
        "nfreq": gsdata.nfreqs,
        "freq_min": gsdata.freqs.min().to_value("MHz"),
        "freq_max": gsdata.freqs.max().to_value("MHz"),
        "freq_res": (gsdata.freqs[1] - gsdata.freqs[0]).to_value("MHz"),
    }

    encode(outfile, p=gsdata.data[:, 0], meta=meta, ancillary=ancillary)
